#!/usr/bin/env  python3

import logging
from time import sleep
from typing import List
import yaml
from yaml.loader import SafeLoader
from collections import namedtuple
from git import Repo
from .job import Job

WorkflowStats = namedtuple('WorkflowStarts', ['success', 'failure'])

class WorkflowManager:
    """
    Circleci workflow manager
    """
    def __init__(self, circleci_config: str, exclude: List[str] = None):
        """
        Parameters
        ----------
        circleci_config: str
            Circleci configuration file
        exclude: List[str]
            List of workflows to exclude (default: NONE)
        """
        self.circleci_config = circleci_config
        self.exclude = exclude if exclude else []
        #self.workflows = {}

    def parse_config(self):
        with open(self.circleci_config, 'r') as config_file:
            config = yaml.load(config_file, Loader=SafeLoader)

        exclude_workflows = self.exclude
        workflows = {}
        for name, workflow in config['workflows'].items():
            try:
                jobs = workflow.get('jobs', None)
                if jobs and name not in exclude_workflows:
                    workflows[name] = jobs
            except AttributeError:
                continue

        self.workflows = {}
        for name, jobs in workflows.items():
            self.workflows[name] = []
            for job in jobs:
                job_name = job
                require = None
                if isinstance(job, dict):
                    job_name, options = list(job.items())[0]
                    require = options.get('requires', None)

                workflow_job = Job(job_name, name, self.circleci_config, require)
                self.workflows[name].append(workflow_job)

    def loop(self, gitpath: str):
        logging.basicConfig(level=logging.INFO, format='[%(asctime)s] %(levelname)s - %(message)s')
        repo = Repo(gitpath)
        commit = repo.commit()
        previous_commit = None
        try:
            while True:
                if commit != previous_commit:
                    msg = f"New commit: {commit.summary}"
                    logging.info(msg)
                    self.run()
                    previous_commit = commit
                sleep(1)
                commit = repo.commit()


        except KeyboardInterrupt:
            logging.error("Canceled loop")

    def run(self):
        """
        Run circle workflows
        """
        #import pdb; pdb.set_trace()
        self.stats = {}
        self.parse_config()
        for workflow, jobs in self.workflows.items():
            print(f"[+] Running {workflow} workflow")
            all_executed = False
            executed = []
            failed = []
            no_executed = jobs
            while not all_executed:
                for job in no_executed:
                    executed, failed, no_executed = self.run_job(job, executed, failed, no_executed)

                if not no_executed:
                    all_executed = True
                    break

            self.stats[workflow] = WorkflowStats(success=executed, failure=failed)

    def report(self):
        """
        Report of job execution stats in each workflow
        """
        title = "EXECUTION STATS"
        line = "="*len(title)
        print(f"{line}\n{title}\n{line}")

        for workflow, stats in self.stats.items():
            print(f"{workflow}(SUCCESS:{len(stats.success)}, FAILURE:{len(stats.failure)})")
            for success_job in stats.success:
                print(f"\t {success_job.name} [SUCCESS]")
            for failure_job in stats.failure:
                print(f"\t {failure_job.name} [FAILURE]")

            print()



    def run_job(self, job, executed, failed, no_executed):
        """
        Run a job of a workflow
        """
        if job.requires:
            for job_name in job.requires:
                # check if any required job was failed
                for failed_job in failed:
                    if failed_job.name == job_name:
                        no_executed.remove(job)
                        failed.append(job)
                        return executed, failed, no_executed

                executed_require_job = False
                for executed_job in executed:
                    if executed_job.name == job_name:
                        executed_require_job = True
                        break

                # required job wasn't executed
                if not executed_require_job:
                    required_job = None
                    for no_executed_job in no_executed:
                        if no_executed_job.name == job_name:
                            required_job = no_executed_job
                            break

                    if required_job:
                        executed, failed, no_executed = self.run_job(required_job, executed, failed, no_executed)
                        if required_job.failed:
                            no_executed.remove(job)
                            failed.append(job)
                            return executed, failed, no_executed

        job.run()
        no_executed.remove(job)

        if job.failed:
            failed.append(job)
        else:
            executed.append(job)

        return executed, failed, no_executed
